@@ -1,7 +1,7 @@
 #!/usr/bin/python
 #
 # Urwid container widget classes
-#    Copyright (C) 2004-2008  Ian Ward
+#    Copyright (C) 2004-2011  Ian Ward
 #
 #    This library is free software; you can redistribute it and/or
 #    modify it under the terms of the GNU Lesser General Public
@@ -19,10 +19,13 @@
 #
 # Urwid web site: http://excess.org/urwid/
 
-from util import *
-from widget import *
-from decoration import *
-from command_map import command_map
+from urwid.util import is_mouse_press
+from urwid.widget import Widget, BoxWidget, FlowWidget, Divider
+from urwid.decoration import Padding, Filler, calculate_padding, calculate_filler, \
+    decompose_align_width, decompose_valign_height
+from urwid.monitored_list import MonitoredList
+from urwid.canvas import CompositeCanvas, CanvasOverlay, CanvasCombine, \
+    SolidCanvas, CanvasJoin
 
 
 class WidgetContainer(Widget):
@@ -82,7 +85,7 @@ class WidgetContainer(Widget):
         raise ValueError
 
     def count(self, value):
-        return sum([1 for v in self if v == value])
+        return sum(1 for v in self if v == value)
 
 
 
@@ -118,7 +121,7 @@ class GridFlow(FlowWidget):
         """Set the cell in focus.  
         
         cell -- widget or integer index into self.cells"""
-        if type(cell) == type(0):
+        if type(cell) == int:
             assert cell>=0 and cell<len(self.cells)
             self.focus_cell = self.cells[cell]
         else:
@@ -128,6 +131,10 @@ class GridFlow(FlowWidget):
         self._invalidate()
         
 
+    def get_focus (self):
+        """Return the widget in focus."""
+        return self.focus_cell
+
     def get_display_widget(self, size):
         """
         Arrange the cells into columns (and possibly a pile) for 
@@ -152,14 +159,14 @@ class GridFlow(FlowWidget):
         d = Divider()
         if len(self.cells) == 0: # how dull
             return d
-        
+
         if self.v_sep > 1:
             # increase size of divider
             d.top = self.v_sep-1
-        
+
         # cells per row
-        bpr = (maxcol+self.h_sep) / (self.cell_width+self.h_sep)
-        
+        bpr = (maxcol+self.h_sep) // (self.cell_width+self.h_sep)
+
         if bpr == 0: # too narrow, pile them on top of eachother
             l = [self.cells[0]]
             f = 0
@@ -213,7 +220,7 @@ class GridFlow(FlowWidget):
             # unwrap padding
             w = w._original_widget
         w = w.get_focus()
-        #assert w == self.cells[0], `w, self.cells`
+        #assert w == self.cells[0], repr((w, self.cells))
         self.set_focus(w)
 
     def keypress(self, size, key):
@@ -310,7 +317,7 @@ class Overlay(BoxWidget):
             ('relative', percentage 0=top 100=bottom)
         height -- one of:
             None if top_w is a flow or fixed widget
-            number of rows high 
+            number of rows high
             ('fixed bottom', rows)  Only if valign is 'fixed top'
             ('fixed top', rows)  Only if valign is 'fixed bottom'
             ('relative', percentage of total height)
@@ -318,32 +325,43 @@ class Overlay(BoxWidget):
             when width is not fixed
         min_height -- one of:
             minimum number of rows for the widget when height not fixed
-        
+
         Overlay widgets behave similarly to Padding and Filler widgets
         when determining the size and position of top_w.  bottom_w is
         always rendered the full size available "below" top_w.
         """
         self.__super.__init__()
 
-        at,aa,wt,wa=decompose_align_width(align, width, OverlayError)
-        vt,va,ht,ha=decompose_valign_height(valign,height,OverlayError)
-        
         self.top_w = top_w
         self.bottom_w = bottom_w
-        
+
+        self.set_overlay_parameters(align, width, valign, height,
+            min_width, min_height)
+
+    def set_overlay_parameters(self, align, width, valign, height,
+            min_width=None, min_height=None):
+        """
+        Adjust the overlay size and position parameters.
+
+        See __init__() for a description of the parameters.
+        """
+        at,aa,wt,wa=decompose_align_width(align, width, OverlayError)
+        vt,va,ht,ha=decompose_valign_height(valign,height,OverlayError)
+
         self.align_type, self.align_amount = at, aa
         self.width_type, self.width_amount = wt, wa
         if self.width_type and self.width_type != 'fixed':
             self.min_width = min_width
         else:
             self.min_width = None
-        
+
         self.valign_type, self.valign_amount = vt, va
         self.height_type, self.height_amount = ht, ha
         if self.height_type not in ('fixed', None):
             self.min_height = min_height
         else:
             self.min_height = None
+        self._invalidate()
 
     def selectable(self):
         """Return selectable from top_w."""
@@ -494,6 +512,13 @@ class Frame(BoxWidget):
         self.focus_part = part
         self._invalidate()
 
+    def get_focus (self):
+        """Return the part of the frame that is in focus.
+
+        Will be one of 'header', 'footer' or 'body'.
+        """
+        return self.focus_part
+
     def frame_top_bottom(self, size, focus):
         """Calculate the number of rows for the header and footer.
 
@@ -686,7 +711,7 @@ class Pile(Widget): # either FlowWidget or BoxWidget
         self.item_types = []
         for i in range(len(widget_list)):
             w = widget_list[i]
-            if type(w) != type(()):
+            if type(w) != tuple:
                 self.item_types.append(('weight',1))
             elif w[0] == 'flow':
                 f, widget = w
@@ -699,7 +724,7 @@ class Pile(Widget): # either FlowWidget or BoxWidget
                 self.item_types.append((f,height))
                 w = widget
             else:
-                raise PileError, "widget list item invalid %s" % `w`
+                raise PileError, "widget list item invalid %r" % (w,)
             if focus_item is None and w.selectable():
                 focus_item = i
         self.widget_list.set_modified_callback(self._invalidate)
@@ -712,6 +737,20 @@ class Pile(Widget): # either FlowWidget or BoxWidget
             self.focus_item=None
         self.pref_col = 0
 
+    @property
+    def contents(self):
+        for i, w in enumerate(self.widget_list):
+            try:
+                yield w, self.item_types[i]
+            except IndexError:
+                yield w, ('weight', 1)
+
+    def _get_item_types(self, i):
+        try:
+            return self.item_types[i]
+        except IndexError:
+            return 'weight', 1
+
     def selectable(self):
         """Return True if the focus item is selectable."""
         return self.focus_item is not None and self.focus_item.selectable()
@@ -720,7 +759,7 @@ class Pile(Widget): # either FlowWidget or BoxWidget
         """Set the item in focus.  
         
         item -- widget or integer index"""
-        if type(item) == type(0):
+        if type(item) == int:
             assert item>=0 and item<len(self.widget_list)
             self.focus_item = self.widget_list[item]
         else:
@@ -744,7 +783,7 @@ class Pile(Widget): # either FlowWidget or BoxWidget
         Return a size appropriate for passing to self.widget_list[i]
         """
         maxcol = size[0]
-        f, height = self.item_types[i]
+        f, height = self._get_item_types(i)
         if f=='fixed':
             return (maxcol, height)
         elif f=='weight' and len(size)==2:
@@ -753,7 +792,7 @@ class Pile(Widget): # either FlowWidget or BoxWidget
             return (maxcol, item_rows[i])
         else:
             return (maxcol,)
-                    
+
     def get_item_rows(self, size, focus):
         """
         Return a list of the number of rows used by each widget
@@ -763,24 +802,23 @@ class Pile(Widget): # either FlowWidget or BoxWidget
         maxcol = size[0]
         if len(size)==2:
             remaining = size[1]
-        
+
         l = []
-        
+
         if remaining is None:
             # pile is a flow widget
-            for (f, height), w in zip(
-                self.item_types, self.widget_list):
+            for w, (f, height) in self.contents:
                 if f == 'fixed':
-                    l.append( height )
+                    l.append(height)
                 else:
-                    l.append( w.rows( (maxcol,), focus=focus
-                        and self.focus_item == w ))
+                    l.append(w.rows((maxcol,), focus=focus
+                        and self.focus_item == w))
             return l
-            
+
         # pile is a box widget
         # do an extra pass to calculate rows for each widget
         wtotal = 0
-        for (f, height), w in zip(self.item_types, self.widget_list):
+        for w, (f, height) in self.contents:
             if f == 'flow':
                 rows = w.rows((maxcol,), focus=focus and
                     self.focus_item == w )
@@ -799,18 +837,16 @@ class Pile(Widget): # either FlowWidget or BoxWidget
         if remaining < 0: 
             remaining = 0
 
-        i = 0
-        for (f, height), li in zip(self.item_types, l):
+        for i, (w, (f, height)) in enumerate(self.contents):
+            li = l[i]
             if li is None:
                 rows = int(float(remaining)*height
                     /wtotal+0.5)
                 l[i] = rows
                 remaining -= rows
                 wtotal -= height
-            i += 1
         return l
-        
-    
+
     def render(self, size, focus=False):
         """
         Render all widgets in self.widget_list and return the results
@@ -818,10 +854,9 @@ class Pile(Widget): # either FlowWidget or BoxWidget
         """
         maxcol = size[0]
         item_rows = None
-        
+
         combinelist = []
-        i = 0
-        for (f, height), w in zip(self.item_types, self.widget_list):
+        for i, (w, (f, height)) in enumerate(self.contents):
             item_focus = self.focus_item == w
             canv = None
             if f == 'fixed':
@@ -840,11 +875,15 @@ class Pile(Widget): # either FlowWidget or BoxWidget
                         focus=focus and    item_focus )
             if canv:
                 combinelist.append((canv, i, item_focus))
-            i+=1
         if not combinelist:
             return SolidCanvas(" ", size[0], (size[1:]+(0,))[0])
 
-        return CanvasCombine(combinelist)
+        out = CanvasCombine(combinelist)
+        if len(size)==2 and size[1] < out.rows():
+            # flow/fixed widgets rendered too large
+            out = CompositeCanvas(out)
+            out.pad_trim_top_bottom(0, size[1] - out.rows())
+        return out
     
     def get_cursor_coords(self, size):
         """Return the cursor coordinates of the focus widget."""
@@ -854,7 +893,7 @@ class Pile(Widget): # either FlowWidget or BoxWidget
             return None
         
         i = self.widget_list.index(self.focus_item)
-        f, height = self.item_types[i]
+        f, height = self._get_item_types(i)
         item_rows = None
         maxcol = size[0]
         if f == 'fixed' or (f=='weight' and len(size)==2):
@@ -879,61 +918,57 @@ class Pile(Widget): # either FlowWidget or BoxWidget
             for r in item_rows[:i]:
                 y += r
         return x, y
-        
-    
+
     def rows(self, size, focus=False ):
         """Return the number of rows required for this widget."""
         return sum(self.get_item_rows(size, focus))
 
-
     def keypress(self, size, key ):
         """Pass the keypress to the widget in focus.
         Unhandled 'up' and 'down' keys may cause a focus change."""
 
-        maxcol = size[0]
         item_rows = None
         if len(size)==2:
             item_rows = self.get_item_rows( size, focus=True )
 
         i = self.widget_list.index(self.focus_item)
-        f, height = self.item_types[i]
+        f, height = self._get_item_types(i)
         if self.focus_item.selectable():
             tsize = self.get_item_size(size,i,True,item_rows)
             key = self.focus_item.keypress( tsize, key )
-            if command_map[key] not in ('cursor up', 'cursor down'):
+            if self._command_map[key] not in ('cursor up', 'cursor down'):
                 return key
 
-        if command_map[key] == 'cursor up':
+        if self._command_map[key] == 'cursor up':
             candidates = range(i-1, -1, -1) # count backwards to 0
-        else: # command_map[key] == 'cursor down'
+        else: # self._command_map[key] == 'cursor down'
             candidates = range(i+1, len(self.widget_list))
-        
+
         if not item_rows:
             item_rows = self.get_item_rows( size, focus=True )
-    
+
         for j in candidates:
             if not self.widget_list[j].selectable():
                 continue
-            
+
             self._update_pref_col_from_focus(size)
-            old_focus = self.focus_item
             self.set_focus(j)
             if not hasattr(self.focus_item,'move_cursor_to_coords'):
                 return
 
-            f, height = self.item_types[j]
+            f, height = self._get_item_types(i)
             rows = item_rows[j]
-            if command_map[key] == 'cursor up':
+            if self._command_map[key] == 'cursor up':
                 rowlist = range(rows-1, -1, -1)
-            else: # command_map[key] == 'cursor down'
+            else: # self._command_map[key] == 'cursor down'
                 rowlist = range(rows)
             for row in rowlist:
                 tsize=self.get_item_size(size,j,True,item_rows)
                 if self.focus_item.move_cursor_to_coords(
                         tsize,self.pref_col,row):
                     break
-            return                    
-                
+            return
+
         # nothing to select
         return key
 
@@ -1007,7 +1042,7 @@ class Pile(Widget): # either FlowWidget or BoxWidget
 class ColumnsError(Exception):
     pass
 
-        
+
 class Columns(Widget): # either FlowWidget or BoxWidget
     def __init__(self, widget_list, dividechars=0, focus_column=None,
         min_width=1, box_columns=None):
@@ -1016,16 +1051,18 @@ class Columns(Widget): # either FlowWidget or BoxWidget
         dividechars -- blank characters between columns
         focus_column -- index into widget_list of column in focus,
             if None the first selectable widget will be chosen.
-        min_width -- minimum width for each column before it is hidden
+        min_width -- minimum width for each column which is not
+            designated as flow widget in widget_list.
         box_columns -- a list of column indexes containing box widgets
-            whose maxrow is set to the maximum of the rows 
+            whose maxrow is set to the maximum of the rows
             required by columns not listed in box_columns.
 
         widget_list may also contain tuples such as:
+        ('flow', widget) always treat widget as a flow widget
         ('fixed', width, widget) give this column a fixed width
         ('weight', weight, widget) give this column a relative weight
 
-        widgets not in a tuple are the same as ('weight', 1, widget)    
+        widgets not in a tuple are the same as ('weight', 1, widget)
 
         box_columns is ignored when this widget is being used as a
         box widget because in that case all columns are treated as box
@@ -1034,22 +1071,29 @@ class Columns(Widget): # either FlowWidget or BoxWidget
         self.__super.__init__()
         self.widget_list = MonitoredList(widget_list)
         self.column_types = []
+        self.has_flow_type = False
         for i in range(len(widget_list)):
             w = widget_list[i]
-            if type(w) != type(()):
+            if type(w) != tuple:
                 self.column_types.append(('weight',1))
+            elif w[0] == 'flow':
+                f, widget = w
+                self.widget_list[i] = widget
+                self.column_types.append((f,None))
+                self.has_flow_type = True
+                w = widget
             elif w[0] in ('fixed', 'weight'):
                 f,width,widget = w
                 self.widget_list[i] = widget
                 self.column_types.append((f,width))
                 w = widget
             else:
-                raise ColumnsError, "widget list item invalid: %s" % `w`
+                raise ColumnsError, "widget list item invalid: %r" % (w,)
             if focus_column is None and w.selectable():
                 focus_column = i
-                
+
         self.widget_list.set_modified_callback(self._invalidate)
-        
+
         self.dividechars = dividechars
         if focus_column is None:
             focus_column = 0
@@ -1058,7 +1102,15 @@ class Columns(Widget): # either FlowWidget or BoxWidget
         self.min_width = min_width
         self.box_columns = box_columns
         self._cache_maxcol = None
-    
+
+    @property
+    def contents(self):
+        for i, w in enumerate(self.widget_list):
+            try:
+                yield w, self.column_types[i]
+            except IndexError:
+                yield w, ('weight', 1)
+
     def _invalidate(self):
         self._cache_maxcol = None
         self.__super._invalidate()
@@ -1076,7 +1128,7 @@ class Columns(Widget): # either FlowWidget or BoxWidget
         """Set the item in focus.  
         
         item -- widget or integer index"""
-        if type(item) == type(0):
+        if type(item) == int:
             assert item>=0 and item<len(self.widget_list)
             position = item
         else:
@@ -1088,46 +1140,37 @@ class Columns(Widget): # either FlowWidget or BoxWidget
         """Return the widget in focus."""
         return self.widget_list[self.focus_col]
 
-    def column_widths( self, size ):
+    def column_widths(self, size, focus=False):
         """Return a list of column widths.
 
         size -- (maxcol,) if self.widget_list contains flow widgets or
             (maxcol, maxrow) if it contains box widgets.
         """
         maxcol = size[0]
-        if maxcol == self._cache_maxcol:
+        if maxcol == self._cache_maxcol and not self.has_flow_type:
             return self._cache_column_widths
 
-        col_types = self.column_types
-        # hack to support old practice of editing self.widget_list
-        # directly
-        lwl, lct = len(self.widget_list), len(self.column_types)
-        if lwl > lct:
-            col_types = col_types + [('weight',1)] * (lwl-lct)
-            
         widths=[]
-        
+
         weighted = []
         shared = maxcol + self.dividechars
-        growable = 0
-        
-        i = 0
-        for t, width in col_types:
+
+        for i, (w, (t, width)) in enumerate(self.contents):
             if t == 'fixed':
                 static_w = width
+            elif t == 'flow':
+                static_w = w.pack((maxcol,), focus)[0]
             else:
                 static_w = self.min_width
-                
+
             if shared < static_w + self.dividechars:
                 break
-        
-            widths.append( static_w )    
+
+            widths.append(static_w)
             shared -= static_w + self.dividechars
-            if t != 'fixed':
-                weighted.append( (width,i) )
-        
-            i += 1
-        
+            if t not in ('fixed', 'flow'):
+                weighted.append((width,i))
+
         if shared:
             # divide up the remaining space between weighted cols
             weighted.sort()
@@ -1139,18 +1182,18 @@ class Columns(Widget): # either FlowWidget or BoxWidget
                 widths[i] = width
                 grow -= width
                 wtotal -= weight
-        
+
         self._cache_maxcol = maxcol
         self._cache_column_widths = widths
         return widths
-    
+
     def render(self, size, focus=False):
         """Render columns and return canvas.
 
         size -- (maxcol,) if self.widget_list contains flow widgets or
             (maxcol, maxrow) if it contains box widgets.
         """
-        widths = self.column_widths( size )
+        widths = self.column_widths(size, focus)
         if not widths:
             return SolidCanvas(" ", size[0], (size[1:]+(1,))[0])
         
@@ -1170,6 +1213,11 @@ class Columns(Widget): # either FlowWidget or BoxWidget
         l = []
         for i in range(len(widths)):
             mc = widths[i]
+
+            # if the widget has a width of 0, hide it
+            if mc <= 0:
+                continue
+
             w = self.widget_list[i]
             if box_maxrow and i in self.box_columns:
                 sub_size = (mc, box_maxrow)
@@ -1220,6 +1268,8 @@ class Columns(Widget): # either FlowWidget or BoxWidget
             w = self.widget_list[i]
             end = x + widths[i]
             if w.selectable():
+                # sometimes, col == 'left' - that doesn't seem like its handled here, does it?
+                # assert isinstance(x, int) and isinstance(col, int), (x, col)
                 if x > col and best is None:
                     # no other choice
                     best = i, x, end
@@ -1238,7 +1288,7 @@ class Columns(Widget): # either FlowWidget or BoxWidget
         i, x, end = best
         w = self.widget_list[i]
         if hasattr(w,'move_cursor_to_coords'):
-            if type(col)==type(0):
+            if type(col)==int:
                 move_x = min(max(0,col-x),end-x-1)
             else:
                 move_x = col
@@ -1293,13 +1343,13 @@ class Columns(Widget): # either FlowWidget or BoxWidget
         col = None
         if hasattr(w,'get_pref_col'):
             col = w.get_pref_col((widths[self.focus_col],)+size[1:])
-            if type(col)==type(0):
+            if type(col)==int:
                 col += self.focus_col * self.dividechars
                 col += sum( widths[:self.focus_col] )
         if col is None:
             col = self.pref_col
         if col is None and w.selectable():
-            col = widths[self.focus_col]/2
+            col = widths[self.focus_col] // 2
             col += self.focus_col * self.dividechars
             col += sum( widths[:self.focus_col] )
         return col
@@ -1307,7 +1357,7 @@ class Columns(Widget): # either FlowWidget or BoxWidget
     def rows(self, size, focus=0 ):
         """Return the number of rows required by the columns.
         Only makes sense if self.widget_list contains flow widgets."""
-        widths = self.column_widths(size)
+        widths = self.column_widths(size, focus)
     
         rows = 1
         for i in range(len(widths)):
@@ -1334,15 +1384,15 @@ class Columns(Widget): # either FlowWidget or BoxWidget
         i = self.focus_col
         mc = widths[i]
         w = self.widget_list[i]
-        if command_map[key] not in ('cursor up', 'cursor down',
+        if self._command_map[key] not in ('cursor up', 'cursor down',
             'cursor page up', 'cursor page down'):
             self.pref_col = None
         key = w.keypress( (mc,)+size[1:], key )
         
-        if command_map[key] not in ('cursor left', 'cursor right'):
+        if self._command_map[key] not in ('cursor left', 'cursor right'):
             return key
 
-        if command_map[key] == 'cursor left':
+        if self._command_map[key] == 'cursor left':
             candidates = range(i-1, -1, -1) # count backwards to 0
         else: # key == 'right'
             candidates = range(i+1, len(widths))
