@@ -1,7 +1,7 @@
 #!/usr/bin/python
 #
 # Urwid widget decoration classes
-#    Copyright (C) 2004-2010  Ian Ward
+#    Copyright (C) 2004-2011  Ian Ward
 #
 #    This library is free software; you can redistribute it and/or
 #    modify it under the terms of the GNU Lesser General Public
@@ -20,9 +20,13 @@
 # Urwid web site: http://excess.org/urwid/
 
 
-from util import *
-from widget import *
-from split_repr import remove_defaults
+from urwid.util import int_scale
+from urwid.widget import Widget, WidgetError, \
+    BOX, FLOW, LEFT, CENTER, RIGHT, PACK, CLIP, GIVEN, RELATIVE, RELATIVE_100, \
+    delegate_to_widget_mixin
+from urwid.split_repr import remove_defaults
+from urwid.canvas import CompositeCanvas, SolidCanvas
+from urwid.widget import Divider, Edit, Text, SolidFill # doctests
 
 
 class WidgetDecoration(Widget):  # "decorator" was already taken
@@ -38,7 +42,7 @@ class WidgetDecoration(Widget):  # "decorator" was already taken
 
         Don't actually do this -- use a WidgetDecoration subclass
         instead, these are not real widgets:
-        >>> WidgetDecoration(Text("hi"))
+        >>> WidgetDecoration(Text(u"hi"))
         <WidgetDecoration flow widget <Text flow widget 'hi'>>
         """
         self._original_widget = original_widget
@@ -80,41 +84,54 @@ class WidgetDecoration(Widget):  # "decorator" was already taken
         return self._original_widget.sizing()
 
 
+class WidgetPlaceholder(delegate_to_widget_mixin('_original_widget'),
+        WidgetDecoration):
+    """
+    This is a do-nothing decoration widget that can be used for swapping
+    between widgets without modifying the container of this widget.
+
+    This can be useful for making an interface with a number of distinct
+    pages or for showing and hiding menu or status bars.
+
+    The widget displayed is stored as a .original_widget property and
+    can be changed by assigning a new widget to it.
+    """
+    pass
+
+
 class AttrMapError(WidgetError):
     pass
 
-class AttrMap(WidgetDecoration):
+class AttrMap(delegate_to_widget_mixin('_original_widget'), WidgetDecoration):
     """
     AttrMap is a decoration that maps one set of attributes to another for
     a FlowWidget or BoxWidget
     """
-    no_cache = ["rows"]
-
     def __init__(self, w, attr_map, focus_map=None):
         """
         w -- widget to wrap (stored as self.original_widget)
         attr_map -- attribute to apply to w, or dictionary of attribute mappings
         focus_map -- attribute to apply when in focus or dictionary of
             attribute mappings, if None use attr
-        
+
         This object will pass all function calls and variable references
         to the wrapped widget.
 
-        >>> AttrMap(Divider("!"), 'bright')
-        <AttrMap flow widget <Divider flow widget div_char='!'> attr_map={None: 'bright'}>
+        >>> AttrMap(Divider(u"!"), 'bright')
+        <AttrMap flow widget <Divider flow widget '!'> attr_map={None: 'bright'}>
         >>> AttrMap(Edit(), 'notfocus', 'focus')
         <AttrMap selectable flow widget <Edit selectable flow widget '' edit_pos=0> attr_map={None: 'notfocus'} focus_map={None: 'focus'}>
         >>> size = (5,)
-        >>> am = AttrMap(Text("hi"), 'greeting', 'fgreet')
-        >>> am.render(size, focus=False).content().next()
-        [('greeting', None, 'hi   ')]
+        >>> am = AttrMap(Text(u"hi"), 'greeting', 'fgreet')
+        >>> am.render(size, focus=False).content().next() # ... = b in Python 3
+        [('greeting', None, ...'hi   ')]
         >>> am.render(size, focus=True).content().next()
-        [('fgreet', None, 'hi   ')]
-        >>> am2 = AttrMap(Text(('word', "hi")), {None:'bg', 'word':'greeting'})
+        [('fgreet', None, ...'hi   ')]
+        >>> am2 = AttrMap(Text(('word', u"hi")), {'word':'greeting', None:'bg'})
         >>> am2
-        <AttrMap flow widget <Text flow widget 'hi'> attr_map={None: 'bg', 'word': 'greeting'}>
+        <AttrMap flow widget <Text flow widget 'hi'> attr_map={'word': 'greeting', None: 'bg'}>
         >>> am2.render(size).content().next()
-        [('greeting', None, 'hi'), ('bg', None, '   ')]
+        [('greeting', None, ...'hi'), ('bg', None, ...'   ')]
         """
         self.__super.__init__(w)
 
@@ -203,16 +220,6 @@ class AttrMap(WidgetDecoration):
         canv = CompositeCanvas(canv)
         canv.fill_attr_apply(attr_map)
         return canv
-    
-    # just use our original widget's methods
-    selectable = property(lambda self:self._original_widget.selectable)
-    get_cursor_coords = property(lambda self:self._original_widget.get_cursor_coords)
-    get_pref_col = property(lambda self:self._original_widget.get_pref_col)
-    keypress = property(lambda self:self._original_widget.keypress)
-    move_cursor_to_coords = property(lambda self:self._original_widget.move_cursor_to_coords)
-    rows = property(lambda self:self._original_widget.rows)
-    mouse_event = property(lambda self:self._original_widget.mouse_event)
-    sizing = property(lambda self:self._original_widget.sizing)
 
 
 
@@ -222,25 +229,25 @@ class AttrWrap(AttrMap):
         w -- widget to wrap (stored as self.original_widget)
         attr -- attribute to apply to w
         focus_attr -- attribute to apply when in focus, if None use attr
-        
+
         This widget is a special case of the new AttrMap widget, and it
-        will pass all function calls and variable references to the wrapped 
+        will pass all function calls and variable references to the wrapped
         widget.  This class is maintained for backwards compatibility only,
         new code should use AttrMap instead.
 
-        >>> AttrWrap(Divider("!"), 'bright')
-        <AttrWrap flow widget <Divider flow widget div_char='!'> attr='bright'>
+        >>> AttrWrap(Divider(u"!"), 'bright')
+        <AttrWrap flow widget <Divider flow widget '!'> attr='bright'>
         >>> AttrWrap(Edit(), 'notfocus', 'focus')
         <AttrWrap selectable flow widget <Edit selectable flow widget '' edit_pos=0> attr='notfocus' focus_attr='focus'>
         >>> size = (5,)
-        >>> aw = AttrWrap(Text("hi"), 'greeting', 'fgreet')
+        >>> aw = AttrWrap(Text(u"hi"), 'greeting', 'fgreet')
         >>> aw.render(size, focus=False).content().next()
-        [('greeting', None, 'hi   ')]
+        [('greeting', None, ...'hi   ')]
         >>> aw.render(size, focus=True).content().next()
-        [('fgreet', None, 'hi   ')]
+        [('fgreet', None, ...'hi   ')]
         """
         self.__super.__init__(w, attr, focus_attr)
-    
+
     def _repr_attrs(self):
         # only include the focus_attr when it takes effect (not None)
         d = dict(self.__super._repr_attrs(), attr=self.attr)
@@ -322,7 +329,7 @@ class BoxAdapter(WidgetDecoration):
         box_widget -- box widget (stored as self.original_widget)
         height -- number of rows for box widget
 
-        >>> BoxAdapter(SolidFill("x"), 5) # 5-rows of x's
+        >>> BoxAdapter(SolidFill(u"x"), 5) # 5-rows of x's
         <BoxAdapter flow widget <SolidFill box widget 'x'> height=5>
         """
         if hasattr(box_widget, 'sizing') and BOX not in box_widget.sizing():
@@ -346,7 +353,7 @@ class BoxAdapter(WidgetDecoration):
         """
         Return the predetermined height (behave like a flow widget)
         
-        >>> BoxAdapter(SolidFill("x"), 5).rows((20,))
+        >>> BoxAdapter(SolidFill(u"x"), 5).rows((20,))
         5
         """
         return self.height
@@ -427,28 +434,28 @@ class Padding(WidgetDecoration):
         'left' then self.original_widget may be clipped on the right.
 
         >>> size = (7,)
-        >>> Padding(Text("Head"), ('relative', 20)).render(size).text
-        [' Head  ']
-        >>> Padding(Divider("-"), left=2, right=1).render(size).text
-        ['  ---- ']
-        >>> Padding(Divider("*"), 'center', 3).render(size).text
-        ['  ***  ']
-        >>> p=Padding(Text("1234"), 'left', 2, None, 1, 1)
+        >>> Padding(Text(u"Head"), ('relative', 20)).render(size).text # ... = b in Python 3
+        [...' Head  ']
+        >>> Padding(Divider(u"-"), left=2, right=1).render(size).text
+        [...'  ---- ']
+        >>> Padding(Divider(u"*"), 'center', 3).render(size).text
+        [...'  ***  ']
+        >>> p=Padding(Text(u"1234"), 'left', 2, None, 1, 1)
         >>> p
         <Padding flow widget <Text flow widget '1234'> left=1 right=1 width=2>
         >>> p.render(size).text   # align against left
-        [' 12    ', ' 34    ']
+        [...' 12    ', ...' 34    ']
         >>> p.align = 'right'
         >>> p.render(size).text   # align against right
-        ['    12 ', '    34 ']
-        >>> Padding(Text("hi\nthere"), 'right').render(size).text
-        ['  hi   ', '  there']
+        [...'    12 ', ...'    34 ']
+        >>> Padding(Text(u"hi\nthere"), 'right').render(size).text
+        [...'  hi   ', ...'  there']
 
         """
         self.__super.__init__(w)
 
         # convert obsolete parameters 'fixed left' and 'fixed right':
-        if type(align) == type(()) and align[0] in ('fixed left', 
+        if type(align) == tuple and align[0] in ('fixed left', 
             'fixed right'):
             if align[0]=='fixed left':
                 left = align[1]
@@ -456,7 +463,7 @@ class Padding(WidgetDecoration):
             else:
                 right = align[1]
                 align = RIGHT
-        if type(width) == type(()) and width[0] in ('fixed left', 
+        if type(width) == tuple and width[0] in ('fixed left', 
             'fixed right'):
             if width[0]=='fixed left':
                 left = width[1]
@@ -544,7 +551,7 @@ class Padding(WidgetDecoration):
         Override this method to define custom padding behaviour."""
         maxcol = size[0]
         if self._width_type == CLIP:
-            width, ignore = self._original_widget.pack(focus=focus)
+            width, ignore = self._original_widget.pack((), focus=focus)
             return calculate_left_right_padding(maxcol,
                 self._align_type, self._align_amount, 
                 CLIP, width, None, self.left, self.right)
@@ -605,7 +612,7 @@ class Padding(WidgetDecoration):
         left, right = self.padding_values(size, True)
         maxcol = size[0]
         maxvals = (maxcol-left-right,)+size[1:] 
-        if type(x)==type(0):
+        if type(x)==int:
             if x < left: 
                 x = left
             elif x >= maxcol-right: 
@@ -634,7 +641,7 @@ class Padding(WidgetDecoration):
         maxcol = size[0]
         maxvals = (maxcol-left-right,)+size[1:] 
         x = self._original_widget.get_pref_col(maxvals)
-        if type(x) == type(0):
+        if type(x) == int:
             return x+left
         return x
         
@@ -797,7 +804,7 @@ class Filler(WidgetDecoration):
             return False
         
         top, bottom = self.filler_values(size, True)
-        if row < top or row >= maxcol-bottom:
+        if row < top or row >= maxrow-bottom:
             return False
 
         if self.height_type is None:
@@ -806,7 +813,6 @@ class Filler(WidgetDecoration):
         return self._original_widget.mouse_event((maxcol, maxrow-top-bottom), 
             event, button,col, row-top, focus)
 
-        
 def normalize_align(align, err):
     """
     Split align into (align_type, align_amount).  Raise exception err
@@ -814,11 +820,11 @@ def normalize_align(align, err):
     """
     if align in (LEFT, CENTER, RIGHT):
         return (align, 0)
-    elif type(align) == type(()) and len(align) == 2 and align[0] == RELATIVE:
+    elif type(align) == tuple and len(align) == 2 and align[0] == RELATIVE:
         return align
-    raise err("align value %s is not one of 'left', 'center', "
+    raise err("align value %r is not one of 'left', 'center', "
         "'right', ('relative', percentage 0=left 100=right)" 
-        % `align`)
+        % (align,))
 
 def simplify_align(align_type, align_amount):
     """
@@ -836,13 +842,13 @@ def normalize_width(width, err):
     """
     if width in (CLIP, PACK):
         return (width, 0)
-    elif type(width) == type(0):
+    elif type(width) == int:
         return (GIVEN, width)
-    elif type(width) == type(()) and len(width) == 2 and width[0] == RELATIVE:
+    elif type(width) == tuple and len(width) == 2 and width[0] == RELATIVE:
         return width
-    raise err("width value %s is not one of fixed number of columns, "
+    raise err("width value %r is not one of fixed number of columns, "
         "'pack', ('relative', percentage of total width), 'clip'" 
-        % `width`)
+        % (width,))
 
 def simplify_width(width_type, width_amount):
     """
@@ -863,24 +869,24 @@ def decompose_align_width( align, width, err ):
         align_type, align_amount = align
         assert align_type in ('left','center','right','fixed left',
             'fixed right','relative')
-    except:
-        raise err("align value %s is not one of 'left', 'center', "
+    except (AssertionError, ValueError, TypeError):
+        raise err("align value %r is not one of 'left', 'center', "
             "'right', ('fixed left', columns), ('fixed right', "
             "columns), ('relative', percentage 0=left 100=right)" 
-            % `align`)
+            % (align,))
 
     try:
         if width is None:
             width = None, None
-        elif type(width) == type(0):
+        elif type(width) == int:
             width = 'fixed', width
         width_type, width_amount = width
         assert width_type in ('fixed','fixed right','fixed left',
             'relative', None)
-    except:
-        raise err("width value %s is not one of ('fixed', columns "
+    except (AssertionError, ValueError, TypeError):
+        raise err("width value %r is not one of ('fixed', columns "
             "width), ('fixed right', columns), ('relative', "
-            "percentage of total width), None" % `width`)
+            "percentage of total width), None" % (width,))
         
     if width_type == 'fixed left' and align_type != 'fixed right':
         raise err("fixed left width may only be used with fixed "
@@ -898,18 +904,18 @@ def decompose_valign_height( valign, height, err ):
             valign = (valign,0)
         valign_type, valign_amount = valign
         assert valign_type in ('top','middle','bottom','fixed top','fixed bottom','relative')
-    except:
-        raise err, "Invalid valign: %s" % `valign`
+    except (AssertionError, ValueError, TypeError):
+        raise err, "Invalid valign: %r" % (valign,)
 
     try:
         if height is None:
             height = None, None
-        elif type(height) == type(0):
+        elif type(height) == int:
             height=('fixed',height)
         height_type, height_amount = height
         assert height_type in (None, 'fixed','fixed bottom','fixed top','relative')
-    except:
-        raise err, "Invalid height: %s"%`height`
+    except (AssertionError, ValueError, TypeError):
+        raise err, "Invalid height: %r"%(height,)
         
     if height_type == 'fixed top' and valign_type != 'fixed bottom':
         raise err, "fixed top height may only be used with fixed bottom valign"
@@ -924,7 +930,7 @@ def calculate_filler( valign_type, valign_amount, height_type, height_amount,
     if height_type == 'fixed':
         height = height_amount
     elif height_type == 'relative':
-        height = int(height_amount*maxrow/100+.5)
+        height = int(height_amount*maxrow // 100)
         if min_height is not None:
                 height = max(height, min_height)
     else:
@@ -950,7 +956,7 @@ def calculate_filler( valign_type, valign_amount, height_type, height_amount,
         # need to shrink bottom
         return 0, maxrow-height        
     elif valign_type == 'relative':
-        top = int( (maxrow-height)*valign_amount/100+.5 )
+        top = int( (maxrow-height)*valign_amount // 100)
     elif valign_type == 'bottom':
         top = maxrow-height    
     elif valign_type == 'middle':
