@@ -2,7 +2,7 @@
 # -*- coding: utf-8 -*-
 #
 # Urwid graphics widgets
-#    Copyright (C) 2004-2007  Ian Ward
+#    Copyright (C) 2004-2011  Ian Ward
 #
 #    This library is free software; you can redistribute it and/or
 #    modify it under the terms of the GNU Lesser General Public
@@ -20,14 +20,15 @@
 #
 # Urwid web site: http://excess.org/urwid/
 
-from __future__ import nested_scopes
-
-from util import *
-from canvas import *
-from widget import *
-from container import *
-from escape import utf8decode
-from display_common import AttrSpec
+from urwid.util import decompose_tagmarkup, get_encoding_mode
+from urwid.canvas import CompositeCanvas, CanvasJoin, TextCanvas, \
+    CanvasCombine, SolidCanvas
+from urwid.widget import WidgetMeta, BoxWidget, FlowWidget, FixedWidget, \
+    nocache_widget_render, nocache_widget_render_instance, fixed_size, \
+    WidgetWrap, Divider, SolidFill, Text, CENTER, CLIP
+from urwid.container import Pile, Columns
+from urwid.display_common import AttrSpec
+from urwid.decoration import WidgetDecoration, AttrWrap
 
 class BigText(FixedWidget):
     def __init__(self, markup, font):
@@ -88,42 +89,74 @@ class BigText(FixedWidget):
             canv = CompositeCanvas(canv)
         canv.set_depends([])
         return canv
-        
+
 
 class LineBox(WidgetDecoration, WidgetWrap):
-    def __init__(self, original_widget):
-        """Draw a line around original_widget."""
-        
-        tlcorner=None; tline=None; lline=None
-        trcorner=None; blcorner=None; rline=None
-        bline=None; brcorner=None
-        
-        def use_attr( a, t ):
-            if a is not None:
-                t = urwid.AttrWrap(t, a)
-            return t
-            
-        tline = use_attr( tline, Divider(utf8decode("─")))
-        bline = use_attr( bline, Divider(utf8decode("─")))
-        lline = use_attr( lline, SolidFill(utf8decode("│")))
-        rline = use_attr( rline, SolidFill(utf8decode("│")))
-        tlcorner = use_attr( tlcorner, Text(utf8decode("┌")))
-        trcorner = use_attr( trcorner, Text(utf8decode("┐")))
-        blcorner = use_attr( blcorner, Text(utf8decode("└")))
-        brcorner = use_attr( brcorner, Text(utf8decode("┘")))
-        top = Columns([ ('fixed', 1, tlcorner),
-            tline, ('fixed', 1, trcorner) ])
-        middle = Columns( [('fixed', 1, lline),
-            original_widget, ('fixed', 1, rline)], box_columns = [0,2],
-            focus_column = 1)
-        bottom = Columns([ ('fixed', 1, blcorner),
-            bline, ('fixed', 1, brcorner) ])
-        pile = Pile([('flow',top),middle,('flow',bottom)],
-            focus_item = 1)
-        
+
+    def __init__(self, original_widget, title="",
+                 tlcorner=u'┌', tline=u'─', lline=u'│',
+                 trcorner=u'┐', blcorner=u'└', rline=u'│',
+                 bline=u'─', brcorner=u'┘'):
+        """
+        Draw a line around original_widget.
+
+        Use 'title' to set an initial title text with will be centered
+        on top of the box.
+
+        You can also override the widgets used for the lines/corners:
+            tline: top line
+            bline: bottom line
+            lline: left line
+            rline: right line
+            tlcorner: top left corner
+            trcorner: top right corner
+            blcorner: bottom left corner
+            brcorner: bottom right corner
+
+        """
+
+        tline, bline = Divider(tline), Divider(bline)
+        lline, rline = SolidFill(lline), SolidFill(rline)
+        tlcorner, trcorner = Text(tlcorner), Text(trcorner)
+        blcorner, brcorner = Text(blcorner), Text(brcorner)
+
+        self.title_widget = Text(self.format_title(title))
+        self.tline_widget = Columns([
+            tline,
+            ('flow', self.title_widget),
+            tline,
+        ])
+
+        top = Columns([
+            ('fixed', 1, tlcorner),
+            self.tline_widget,
+            ('fixed', 1, trcorner)
+        ])
+
+        middle = Columns([
+            ('fixed', 1, lline),
+            original_widget,
+            ('fixed', 1, rline),
+        ], box_columns = [0,2], focus_column=1)
+
+        bottom = Columns([
+            ('fixed', 1, blcorner), bline, ('fixed', 1, brcorner)
+        ])
+
+        pile = Pile([('flow', top), middle, ('flow', bottom)], focus_item=1)
+
         WidgetDecoration.__init__(self, original_widget)
         WidgetWrap.__init__(self, pile)
 
+    def format_title(self, text):
+        if len(text) > 0:
+            return " %s " % text
+        else:
+            return ""
+
+    def set_title(self, text):
+        self.title_widget.set_text(self.format_title(text))
+        self.tline_widget._invalidate()
 
 class BarGraphMeta(WidgetMeta):
     """
@@ -158,9 +191,9 @@ class BarGraph(BoxWidget):
     __metaclass__ = BarGraphMeta
     ignore_focus = True
 
-    eighths = utf8decode(" ▁▂▃▄▅▆▇")
-    hlines =  utf8decode("_⎺⎻─⎼⎽")
-    
+    eighths = u' ▁▂▃▄▅▆▇'
+    hlines  = u'_⎺⎻─⎼⎽'
+
     def __init__(self, attlist, hatt=None, satt=None):
         """
         Create a bar graph with the passed display characteristics.
@@ -206,10 +239,10 @@ class BarGraph(BoxWidget):
         self.attr = []
         self.char = []
         if len(attlist) < 2:
-            raise BarGraphError, "attlist must include at least background and seg1: %s" % `attlist`
+            raise BarGraphError, "attlist must include at least background and seg1: %r" % (attlist,)
         assert len(attlist) >= 2, 'must at least specify bg and fg!'
         for a in attlist:
-            if type(a)!=type(()):
+            if type(a)!=tuple:
                 self.attr.append(a)
                 self.char.append(' ')
             else:
@@ -220,7 +253,7 @@ class BarGraph(BoxWidget):
         self.hatt = []
         if hatt is None:
             hatt = [self.attr[0]]
-        elif type(hatt)!=type([]):
+        elif type(hatt)!=list:
             hatt = [hatt]
         self.hatt = hatt
         
@@ -229,12 +262,12 @@ class BarGraph(BoxWidget):
         for i in satt.items():
             try:
                 (fg,bg), attr = i
-            except:
-                raise BarGraphError, "satt not in (fg,bg:attr) form: %s"%`i`
-            if type(fg) != type(0) or fg >= len(attlist):
-                raise BarGraphError, "fg not valid integer: %s"%`fg`
-            if type(bg) != type(0) or bg >= len(attlist):
-                raise BarGraphError, "bg not valid integer: %s"%`fg`
+            except ValueError:
+                raise BarGraphError, "satt not in (fg,bg:attr) form: %r"%(i,)
+            if type(fg) != int or fg >= len(attlist):
+                raise BarGraphError, "fg not valid integer: %r"%(fg,)
+            if type(bg) != int or bg >= len(attlist):
+                raise BarGraphError, "bg not valid integer: %r"%(fg,)
             if fg<=bg:
                 raise BarGraphError, "fg (%s) not > bg (%s)" %(fg,bg)
         self.satt = satt
@@ -399,7 +432,7 @@ class BarGraph(BoxWidget):
         def fill_row( row, chnum ):
             rout = []
             for bar_type, width in row:
-                if (type(bar_type) == type(0) and 
+                if (type(bar_type) == int and 
                         len(self.hatt) > bar_type ):
                     rout.append( ((bar_type, chnum), width))
                     continue
@@ -449,7 +482,7 @@ class BarGraph(BoxWidget):
                 l1 = (bt1, w1-w2)
             elif w2>w1:
                 l2 = (bt2, w2-w1)
-            if type(bt1)==type(()):
+            if type(bt1)==tuple:
                 return (bt1,wmin), l1, l2
             if not self.satt.has_key( (bt2, bt1) ):
                 if r<4:
@@ -491,13 +524,13 @@ class BarGraph(BoxWidget):
             assert r == 0
             # copy whole blocks
             if y_count > 7:
-                o.append( (y_count/8*8 , row) )
+                o.append( (y_count // 8 * 8 , row) )
                 y_count = y_count %8
                 if not y_count:
                     continue
             o.append( (y_count, row) )
             r = y_count
-        return [(y/8, row) for (y,row) in o]
+        return [(y // 8, row) for (y,row) in o]
             
             
     def render(self, size, focus=False):
@@ -511,7 +544,7 @@ class BarGraph(BoxWidget):
         for y_count, row in disp:
             l = []
             for bar_type, width in row:
-                if type(bar_type) == type(()):
+                if type(bar_type) == tuple:
                     if len(bar_type) == 3:
                         # vertical eighths
                         fg,bg,k = bar_type
@@ -565,7 +598,6 @@ def calculate_bargraph_display( bardata, top, bar_widths, maxrow ):
     maxcol = sum(bar_widths)
 
     # build intermediate data structure
-    bars = len(bardata)
     rows = [None]*maxrow
     def add_segment( seg_num, col, row, width, rows=rows ):
         if rows[row]:
@@ -607,7 +639,7 @@ def calculate_bargraph_display( bardata, top, bar_widths, maxrow ):
         col += width
         barnum += 1
     
-    #print `rows`
+    #print repr(rows)
     # build rowsets data structure
     rowsets = []
     y_count = 0
@@ -664,7 +696,7 @@ def calculate_bargraph_display( bardata, top, bar_widths, maxrow ):
             c += ln
             if c == maxcol:
                 break
-            assert i<len(last), `on, maxcol`
+            assert i<len(last), repr((on, maxcol))
             la, ln = last[i]
     
         if i < len(last): 
@@ -754,7 +786,10 @@ def scale_bar_values( bar, top, maxrow ):
 
 
 class ProgressBar( FlowWidget ):
-    eighths = utf8decode(" ▏▎▍▌▋▊▉")
+    eighths = u' ▏▎▍▌▋▊▉'
+
+    text_align = CENTER
+
     def __init__(self, normal, complete, current=0, done=100, satt=None):
         """
         normal -- attribute for uncomplete part of progress bar
@@ -768,33 +803,42 @@ class ProgressBar( FlowWidget ):
         """
         self.normal = normal
         self.complete = complete
-        self.current = current
-        self.done = done
+        self._current = current
+        self._done = done
         self.satt = satt
-    
+
     def set_completion(self, current ):
         """
         current -- current progress
         """
-        self.current = current
+        self._current = current
         self._invalidate()
-    
+    current = property(lambda self:self._current, set_completion)
+
+    def _set_done(self, done):
+        """
+        done -- progress amount at 100%
+        """
+        self._done = done
+        self._invalidate()
+    done = property(lambda self:self._done, _set_done)
+
     def rows(self, size, focus=False):
-        """
-        Return 1.
-        """
         return 1
 
+    def get_text(self):
+        """
+        Return the progress bar percentage text.
+        """
+        percent = min(100, max(0, int(self.current * 100 / self.done)))
+        return str(percent) + " %"
+
     def render(self, size, focus=False):
         """
         Render the progress bar.
         """
         (maxcol,) = size
-        percent = int( self.current*100/self.done )
-        if percent < 0: percent = 0
-        if percent > 100: percent = 100
-            
-        txt=Text( str(percent)+" %", 'center', 'clip' )
+        txt = Text(self.get_text(), self.text_align, CLIP )
         c = txt.render((maxcol,))
 
         cf = float( self.current ) * maxcol / self.done
