@@ -1,7 +1,7 @@
 #!/usr/bin/python
 #
 # Urwid listbox class
-#    Copyright (C) 2004-2010  Ian Ward
+#    Copyright (C) 2004-2011  Ian Ward
 #
 #    This library is free software; you can redistribute it and/or
 #    modify it under the terms of the GNU Lesser General Public
@@ -19,13 +19,13 @@
 #
 # Urwid web site: http://excess.org/urwid/
 
-from util import *
-from canvas import *
-from widget import *
-from decoration import calculate_filler, decompose_valign_height
-import signals
-from monitored_list import MonitoredList
-from command_map import command_map
+from urwid.util import is_mouse_press
+from urwid.canvas import SolidCanvas, CanvasCombine
+from urwid.widget import BoxWidget, nocache_widget_render_instance
+from urwid.decoration import calculate_filler, decompose_valign_height
+from urwid import signals
+from urwid.signals import connect_signal
+from urwid.monitored_list import MonitoredList
 
 
 class ListWalkerError(Exception):
@@ -40,17 +40,23 @@ class ListWalker(object):
 
     def _modified(self):
         signals.emit_signal(self, "modified")
-    
+
 
 class PollingListWalker(object):  # NOT ListWalker subclass
     def __init__(self, contents):
         """
         contents -- list to poll for changes
+
+        This class is deprecated.  Use SimpleListWalker instead.
         """
+        import warnings
+        warnings.warn("PollingListWalker is deprecated, "
+            "use SimpleListWalker instead.", DeprecationWarning)
+
         self.contents = contents
-        if not type(contents) == type([]) and not hasattr( 
+        if not type(contents) == list and not hasattr( 
             contents, '__getitem__' ):
-            raise ListWalkerError, "SimpleListWalker expecting list like object, got: "+`contents`
+            raise ListWalkerError, "SimpleListWalker expecting list like object, got: %r"%(contents,)
         self.focus = 0
     
     def _clamp_focus(self):
@@ -65,7 +71,7 @@ class PollingListWalker(object):  # NOT ListWalker subclass
 
     def set_focus(self, position):
         """Set focus position."""
-        assert type(position) == type(1)
+        assert type(position) == int
         self.focus = position
 
     def get_next(self, start_from):
@@ -94,8 +100,8 @@ class SimpleListWalker(MonitoredList, ListWalker):
         detected automatically and will cause ListBox objects using
         this list walker to be updated.
         """
-        if not type(contents) == type([]) and not hasattr(contents, '__getitem__'):
-            raise ListWalkerError, "SimpleListWalker expecting list like object, got: "+`contents`
+        if not type(contents) == list and not hasattr(contents, '__getitem__'):
+            raise ListWalkerError, "SimpleListWalker expecting list like object, got: %r"%(contents,)
         MonitoredList.__init__(self, contents)
         self.focus = 0
 
@@ -132,7 +138,7 @@ class SimpleListWalker(MonitoredList, ListWalker):
 
     def set_focus(self, position):
         """Set focus position."""
-        assert type(position) == type(1)
+        assert type(position) == int
         self.focus = position
         self._modified()
 
@@ -218,7 +224,7 @@ class ListBox(BoxWidget):
         focus_widget, focus_pos = self.body.get_focus()
         if focus_widget is None: #list box is empty?
             return None,None,None
-        top_pos = bottom_pos = focus_pos
+        top_pos = focus_pos
         
         offset_rows, inset_rows = self.get_focus_offset_inset(
             (maxcol,maxrow))
@@ -275,7 +281,6 @@ class ListBox(BoxWidget):
             next, pos = self.body.get_next( pos )
             if next is None: # run out of widgets below?
                 break
-            bottom_pos = pos
                 
             n_rows = next.rows( (maxcol,) )
             fill_below.append( (next, pos, n_rows) )
@@ -339,17 +344,17 @@ class ListBox(BoxWidget):
         for widget,w_pos,w_rows in fill_above:
             canvas = widget.render((maxcol,))
             if w_rows != canvas.rows():
-                raise ListBoxError, "Widget %s at position %s within listbox calculated %d rows but rendered %d!"% (`widget`,`w_pos`,w_rows, canvas.rows())
+                raise ListBoxError, "Widget %r at position %r within listbox calculated %d rows but rendered %d!"% (widget,w_pos,w_rows, canvas.rows())
             rows += w_rows
             combinelist.append((canvas, w_pos, False))
         
         focus_canvas = focus_widget.render((maxcol,), focus=focus)
 
         if focus_canvas.rows() != focus_rows:
-            raise ListBoxError, "Focus Widget %s at position %s within listbox calculated %d rows but rendered %d!"% (`focus_widget`,`focus_pos`,focus_rows, focus_canvas.rows())
+            raise ListBoxError, "Focus Widget %r at position %r within listbox calculated %d rows but rendered %d!"% (focus_widget,focus_pos,focus_rows, focus_canvas.rows())
         c_cursor = focus_canvas.cursor
         if cursor != c_cursor:
-            raise ListBoxError, "Focus Widget %s at position %s within listbox calculated cursor coords %s but rendered cursor coords %s!" %(`focus_widget`,`focus_pos`,`cursor`,`c_cursor`)
+            raise ListBoxError, "Focus Widget %r at position %r within listbox calculated cursor coords %r but rendered cursor coords %r!" %(focus_widget,focus_pos,cursor,c_cursor)
             
         rows += focus_rows
         combinelist.append((focus_canvas, focus_pos, True))
@@ -357,7 +362,7 @@ class ListBox(BoxWidget):
         for widget,w_pos,w_rows in fill_below:
             canvas = widget.render((maxcol,))
             if w_rows != canvas.rows():
-                raise ListBoxError, "Widget %s at position %s within listbox calculated %d rows but rendered %d!"% (`widget`,`w_pos`,w_rows, canvas.rows())
+                raise ListBoxError, "Widget %r at position %r within listbox calculated %d rows but rendered %d!"% (widget,w_pos,w_rows, canvas.rows())
             rows += w_rows
             combinelist.append((canvas, w_pos, False))
         
@@ -370,17 +375,36 @@ class ListBox(BoxWidget):
             final_canvas.trim_end(trim_bottom)
             rows -= trim_bottom
         
-        assert rows <= maxrow, "Listbox contents too long!  Probably urwid's fault (please report): %s" % `top,middle,bottom`
+        assert rows <= maxrow, "Listbox contents too long!  Probably urwid's fault (please report): %r" % ((top,middle,bottom),)
         
         if rows < maxrow:
             bottom_pos = focus_pos
             if fill_below: bottom_pos = fill_below[-1][1]
-            assert trim_bottom==0 and self.body.get_next(bottom_pos) == (None,None), "Listbox contents too short!  Probably urwid's fault (please report): %s" % `top,middle,bottom`
+            assert trim_bottom==0 and self.body.get_next(bottom_pos) == (None,None), "Listbox contents too short!  Probably urwid's fault (please report): %r" % ((top,middle,bottom),)
             final_canvas.pad_trim_top_bottom(0, maxrow - rows)
 
         return final_canvas
 
 
+    def get_cursor_coords(self, size):
+        (maxcol, maxrow) = size
+        
+        middle, top, bottom = self.calculate_visible(
+            (maxcol, maxrow), True)
+        if middle is None:
+            return None
+
+        offset_inset, _ignore1, _ignore2, _ignore3, cursor = middle
+        if not cursor:
+            return None
+
+        x, y = cursor
+        y += offset_inset
+        if y < 0 or y >= maxrow:
+            return None
+        return (x, y)
+
+
     def set_focus_valign(self, valign):
         """Set the focus widget's display offset and inset.
 
@@ -520,9 +544,9 @@ class ListBox(BoxWidget):
         elif coming_from=='above':
             offset = maxrow-rows
         else:
-            offset = (maxrow-rows)/2
+            offset = (maxrow-rows) // 2
         self.shift_focus((maxcol, maxrow), offset)
-    
+
 
     def shift_focus(self, size, offset_inset):
         """Move the location of the current focus relative to the top.
@@ -538,14 +562,14 @@ class ListBox(BoxWidget):
         
         if offset_inset >= 0:
             if offset_inset >= maxrow:
-                raise ListBoxError, "Invalid offset_inset: %s, only %s rows in list box"% (`offset_inset`, `maxrow`)
+                raise ListBoxError, "Invalid offset_inset: %r, only %r rows in list box"% (offset_inset, maxrow)
             self.offset_rows = offset_inset
             self.inset_fraction = (0,1)
         else:
             target, _ignore = self.body.get_focus()
             tgt_rows = target.rows( (maxcol,), True )
             if offset_inset + tgt_rows <= 0:
-                raise ListBoxError, "Invalid offset_inset: %s, only %s rows in target!" %(`offset_inset`, `tgt_rows`)
+                raise ListBoxError, "Invalid offset_inset: %r, only %r rows in target!" %(offset_inset, tgt_rows)
             self.offset_rows = 0
             self.inset_fraction = (-offset_inset,tgt_rows)
         self._invalidate()
@@ -562,7 +586,7 @@ class ListBox(BoxWidget):
             pref_col = widget.get_pref_col((maxcol,))
         if pref_col is None and hasattr(widget,'get_cursor_coords'):
             coords = widget.get_cursor_coords((maxcol,))
-            if type(coords) == type(()):
+            if type(coords) == tuple:
                 pref_col,y = coords
         if pref_col is not None: 
             self.pref_col = pref_col
@@ -660,7 +684,7 @@ class ListBox(BoxWidget):
             # start from preferred row and move back to closest edge
             (pref_col, pref_row) = cursor_coords
             if pref_row < 0 or pref_row >= tgt_rows:
-                raise ListBoxError, "cursor_coords row outside valid range for target. pref_row:%s target_rows:%s"%(`pref_row`,`tgt_rows`)
+                raise ListBoxError, "cursor_coords row outside valid range for target. pref_row:%r target_rows:%r"%(pref_row,tgt_rows)
 
             if coming_from=='above':
                 attempt_rows = range( pref_row, -1, -1 )
@@ -683,8 +707,8 @@ class ListBox(BoxWidget):
         if offset_rows == 0:
             inum, iden = self.inset_fraction
             if inum < 0 or iden < 0 or inum >= iden:
-                raise ListBoxError, "Invalid inset_fraction: %s"%`self.inset_fraction`
-            inset_rows = focus_rows * inum / iden
+                raise ListBoxError, "Invalid inset_fraction: %r"%(self.inset_fraction,)
+            inset_rows = focus_rows * inum // iden
             assert inset_rows < focus_rows, "urwid inset_fraction error (please report)"
         return offset_rows, inset_rows
 
@@ -745,16 +769,16 @@ class ListBox(BoxWidget):
                 return
         
         # pass off the heavy lifting
-        if command_map[key] == 'cursor up':
+        if self._command_map[key] == 'cursor up':
             return self._keypress_up((maxcol, maxrow))
             
-        if command_map[key] == 'cursor down':
+        if self._command_map[key] == 'cursor down':
             return self._keypress_down((maxcol, maxrow))
 
-        if command_map[key] == 'cursor page up':
+        if self._command_map[key] == 'cursor page up':
             return self._keypress_page_up((maxcol, maxrow))
             
-        if command_map[key] == 'cursor page down':
+        if self._command_map[key] == 'cursor page down':
             return self._keypress_page_down((maxcol, maxrow))
 
         return key
@@ -1001,7 +1025,7 @@ class ListBox(BoxWidget):
         # search within snap_rows then visible region
         search_order = ( range( snap_region_start, len(t))
                 + range( snap_region_start-1, -1, -1 ) )
-        #assert 0, `t, search_order`
+        #assert 0, repr((t, search_order))
         bad_choices = []
         cut_off_selectable_chosen = 0
         for i in search_order:
@@ -1184,7 +1208,7 @@ class ListBox(BoxWidget):
         # search within snap_rows then visible region
         search_order = ( range( snap_region_start, len(t))
                 + range( snap_region_start-1, -1, -1 ) )
-        #assert 0, `t, search_order`
+        #assert 0, repr((t, search_order))
         bad_choices = []
         cut_off_selectable_chosen = 0
         for i in search_order:
