@@ -32,19 +32,16 @@ import sys
 import tty
 import signal
 
-import util
-import escape
-from display_common import *
-import signals
+from urwid import util
+from urwid import escape
+from urwid.display_common import BaseScreen, RealTerminal, \
+    UPDATE_PALETTE_ENTRY, AttrSpec, UNPRINTABLE_TRANS_TABLE, \
+    INPUT_DESCRIPTORS_CHANGED
+from urwid import signals
+from urwid.compat import PYTHON3, bytes, B
 
-try:
-    # python >= 2.4
-    from subprocess import Popen, PIPE
-except ImportError:
-    Popen = None
+from subprocess import Popen, PIPE
 
-# replace control characters with ?'s
-_trans_table = "?"*32+"".join([chr(x) for x in range(32,256)])
 
 class Screen(BaseScreen, RealTerminal):
     def __init__(self):
@@ -62,6 +59,7 @@ class Screen(BaseScreen, RealTerminal):
         self.prev_input_resize = 0
         self.set_input_timeouts()
         self.screen_buf = None
+        self._screen_buf_canvas = None
         self._resized = False
         self.maxrow = None
         self.gpm_mev = None
@@ -70,7 +68,6 @@ class Screen(BaseScreen, RealTerminal):
         self._setup_G1_done = False
         self._rows_used = None
         self._cy = 0
-        self._started = False
         self.bright_is_bold = os.environ.get('TERM',None) != "xterm"
         self._next_timeout = None
         self._term_output_file = sys.stdout
@@ -79,8 +76,6 @@ class Screen(BaseScreen, RealTerminal):
         self._resize_pipe_rd, self._resize_pipe_wr = os.pipe()
         fcntl.fcntl(self._resize_pipe_rd, fcntl.F_SETFL, os.O_NONBLOCK)
 
-    started = property(lambda self: self._started)
-
     def _on_update_palette_entry(self, name, *attrspecs):
         # copy the attribute to a dictionary containing the escape seqences
         self._pal_escape[name] = self._attrspec_to_escape(
@@ -113,7 +108,7 @@ class Screen(BaseScreen, RealTerminal):
 
     def _sigwinch_handler(self, signum, frame):
         if not self._resized:
-            os.write(self._resize_pipe_wr, 'R')
+            os.write(self._resize_pipe_wr, B('R'))
         self._resized = True
         self.screen_buf = None
       
@@ -177,17 +172,21 @@ class Screen(BaseScreen, RealTerminal):
             self._rows_used = None
         else:
             self._rows_used = 0
-        self._old_termios_settings = termios.tcgetattr(0)
+
+        fd = self._term_input_file.fileno()
+        if os.isatty(fd):
+            self._old_termios_settings = termios.tcgetattr(fd)
+            tty.setcbreak(fd)
+
         self.signal_init()
-        tty.setcbreak(self._term_input_file.fileno())
         self._alternate_buffer = alternate_buffer
         self._input_iter = self._run_input_iter()
         self._next_timeout = self.max_wait
         
         if not self._signal_keys_set:
-            self._old_signal_keys = self.tty_signal_keys()
+            self._old_signal_keys = self.tty_signal_keys(fileno=fd)
 
-        self._started = True
+        super(Screen, self).start()
 
     
     def stop(self):
@@ -198,8 +197,12 @@ class Screen(BaseScreen, RealTerminal):
         if not self._started:
             return
         self.signal_restore()
-        termios.tcsetattr(0, termios.TCSADRAIN, 
-            self._old_termios_settings)
+
+        fd = self._term_input_file.fileno()
+        if os.isatty(fd):
+            termios.tcsetattr(fd, termios.TCSADRAIN,
+        self._old_termios_settings)
+
         move_cursor = ""
         if self.gpm_mev:
             self._stop_gpm_tracking()
@@ -216,10 +219,10 @@ class Screen(BaseScreen, RealTerminal):
         self._input_iter = self._fake_input_iter()
 
         if self._old_signal_keys:
-            self.tty_signal_keys(*self._old_signal_keys)
-        
-        self._started = False
-        
+            self.tty_signal_keys(*(self._old_signal_keys + (fd,)))
+
+        super(Screen, self).stop()
+
 
     def run_wrapper(self, fn, alternate_buffer=True):
         """
@@ -411,7 +414,7 @@ class Screen(BaseScreen, RealTerminal):
         ready = None
         fd_list = [self._term_input_file.fileno()]
         if self.gpm_mev is not None:
-            fd_list += [ self.gpm_mev.stdout ]
+            fd_list.append(self.gpm_mev.stdout.fileno())
         while True:
             try:
                 if timeout is None:
@@ -440,11 +443,12 @@ class Screen(BaseScreen, RealTerminal):
     
     def _encode_gpm_event( self ):
         self.gpm_event_pending = False
-        s = self.gpm_mev.stdout.readline()
+        s = self.gpm_mev.stdout.readline().decode('ascii')
         l = s.split(",")
         if len(l) != 6:
             # unexpected output, stop tracking
             self._stop_gpm_tracking()
+            signals.emit_signal(self, INPUT_DESCRIPTORS_CHANGED)
             return []
         ev, x, y, ign, b, m = s.split(",")
         ev = int( ev.split("x")[-1], 16)
@@ -521,7 +525,7 @@ class Screen(BaseScreen, RealTerminal):
                 self._term_output_file.write(escape.DESIGNATE_G1_SPECIAL)
                 self._term_output_file.flush()
                 break
-            except IOError, e:
+            except IOError:
                 pass
         self._setup_G1_done = True
 
@@ -532,6 +536,10 @@ class Screen(BaseScreen, RealTerminal):
 
         assert maxrow == r.rows()
 
+        # quick return if nothing has changed
+        if self.screen_buf and r is self._screen_buf_canvas:
+            return
+
         self._setup_G1()
         
         if self._resized: 
@@ -628,14 +636,21 @@ class Screen(BaseScreen, RealTerminal):
             first = True
             lasta = lastcs = None
             for (a,cs, run) in row:
-                run = run.translate( _trans_table )
+                assert isinstance(run, bytes) # canvases should render with bytes
+                if cs != 'U':
+                    run = run.translate(UNPRINTABLE_TRANS_TABLE)
                 if first or lasta != a:
                     o.append(attr_to_escape(a))
                     lasta = a
                 if first or lastcs != cs:
-                    assert cs in [None, "0"], `cs`
+                    assert cs in [None, "0", "U"], repr(cs)
+                    if lastcs == "U":
+                        o.append( escape.IBMPC_OFF )
+
                     if cs is None:
                         o.append( escape.SI )
+                    elif cs == "U":
+                        o.append( escape.IBMPC_ON )
                     else:
                         o.append( escape.SO )
                     lastcs = cs
@@ -644,9 +659,11 @@ class Screen(BaseScreen, RealTerminal):
             if ins:
                 (inserta, insertcs, inserttext) = ins
                 ias = attr_to_escape(inserta)
-                assert insertcs in [None, "0"], `insertcs`
+                assert insertcs in [None, "0", "U"], repr(insertcs)
                 if cs is None:
                     icss = escape.SI
+                elif cs == "U":
+                    icss = escape.IBMPC_ON
                 else:
                     icss = escape.SO
                 o += [    "\x08"*back, 
@@ -654,6 +671,9 @@ class Screen(BaseScreen, RealTerminal):
                     escape.INSERT_ON, inserttext,
                     escape.INSERT_OFF ]
 
+                if cs == "U":
+                    o.append(escape.IBMPC_OFF)
+
         if r.cursor is not None:
             x,y = r.cursor
             o += [set_cursor_position(x, y), 
@@ -666,7 +686,9 @@ class Screen(BaseScreen, RealTerminal):
         try:
             k = 0
             for l in o:
-                self._term_output_file.write( l )
+                if isinstance(l, bytes) and PYTHON3:
+                    l = l.decode('utf-8')
+                self._term_output_file.write(l)
                 k += len(l)
                 if k > 1024:
                     self._term_output_file.flush()
@@ -678,7 +700,7 @@ class Screen(BaseScreen, RealTerminal):
                 raise
 
         self.screen_buf = sb
-        self.keep_cache_alive_link = r
+        self._screen_buf_canvas = r
                 
     
     def _last_row(self, row):
@@ -761,7 +783,8 @@ class Screen(BaseScreen, RealTerminal):
                 fg = "%d" % (a.foreground_number + 30)
         else:
             fg = "39"
-        st = "1;" * a.bold + "4;" * a.underline + "7;" * a.standout
+        st = ("1;" * a.bold + "4;" * a.underline +
+              "5;" * a.blink + "7;" * a.standout)
         if a.background_high:
             bg = "48;5;%d" % a.background_number
         elif a.background_basic:
@@ -844,7 +867,7 @@ class Screen(BaseScreen, RealTerminal):
 
         modify = ["%d;rgb:%02x/%02x/%02x" % (index, red, green, blue)
             for index, red, green, blue in entries]
-        seq = self._term_output_file.write("\x1b]4;"+";".join(modify)+"\x1b\\")
+        self._term_output_file.write("\x1b]4;"+";".join(modify)+"\x1b\\")
         self._term_output_file.flush()
 
 
