@@ -1,7 +1,7 @@
 #!/usr/bin/python
 #
 # Urwid Window-Icon-Menu-Pointer-style widget classes
-#    Copyright (C) 2004-2008  Ian Ward
+#    Copyright (C) 2004-2011  Ian Ward
 #
 #    This library is free software; you can redistribute it and/or
 #    modify it under the terms of the GNU Lesser General Public
@@ -19,9 +19,15 @@
 #
 # Urwid web site: http://excess.org/urwid/
 
-from widget import *
-from container import *
-from command_map import command_map
+from urwid.widget import Text, WidgetWrap, delegate_to_widget_mixin, BOX
+from urwid.canvas import CompositeCanvas
+from urwid.signals import connect_signal
+from urwid.container import Columns, Overlay
+from urwid.util import is_mouse_press
+from urwid.text_layout import calc_coords
+from urwid.signals import disconnect_signal # doctests
+from urwid.split_repr import python3_repr
+from urwid.decoration import WidgetDecoration
 
 
 class SelectableIcon(Text):
@@ -37,13 +43,13 @@ class SelectableIcon(Text):
         """
         self.__super.__init__(text)
         self._cursor_position = cursor_position
-    
+
     def render(self, size, focus=False):
         """
         Render the text content of this widget with a cursor when
         in focus.
 
-        >>> si = SelectableIcon("[!]")
+        >>> si = SelectableIcon(u"[!]")
         >>> si
         <SelectableIcon selectable flow widget '[!]'>
         >>> si.render((4,), focus=True).cursor
@@ -105,7 +111,7 @@ class CheckBox(WidgetWrap):
         has_mixed -- True if "mixed" is a state to cycle through
         on_state_change, user_data -- shorthand for connect_signal()
             function call for a single callback
-        
+
         Signals supported: 'change'
         Register signal handler with:
           connect_signal(check_box, 'change', callback [,user_data])
@@ -113,15 +119,15 @@ class CheckBox(WidgetWrap):
         Unregister signal handlers with:
           disconnect_signal(check_box, 'change', callback [,user_data])
 
-        >>> CheckBox("Confirm")
+        >>> CheckBox(u"Confirm")
         <CheckBox selectable widget 'Confirm' state=False>
-        >>> CheckBox("Yogourt", "mixed", True)
+        >>> CheckBox(u"Yogourt", "mixed", True)
         <CheckBox selectable widget 'Yogourt' state='mixed'>
-        >>> cb = CheckBox("Extra onions", True)
+        >>> cb = CheckBox(u"Extra onions", True)
         >>> cb
         <CheckBox selectable widget 'Extra onions' state=True>
-        >>> cb.render((20,), focus=True).text  # preview CheckBox
-        ['[X] Extra onions    ']
+        >>> cb.render((20,), focus=True).text # ... = b in Python 3
+        [...'[X] Extra onions    ']
         """
         self.__super.__init__(None) # self.w set by set_state below
         self._label = Text("")
@@ -133,15 +139,15 @@ class CheckBox(WidgetWrap):
             connect_signal(self, 'change', on_state_change, user_data)
         self.set_label(label)
         self.set_state(state)
-    
+
     def _repr_words(self):
         return self.__super._repr_words() + [
-            repr(self.label)]
-    
+            python3_repr(self.label)]
+
     def _repr_attrs(self):
         return dict(self.__super._repr_attrs(),
             state=self.state)
-    
+
     def set_label(self, label):
         """
         Change the check box label.
@@ -149,33 +155,33 @@ class CheckBox(WidgetWrap):
         label -- markup for label.  See Text widget for description
         of text markup.
 
-        >>> cb = CheckBox("foo")
+        >>> cb = CheckBox(u"foo")
         >>> cb
         <CheckBox selectable widget 'foo' state=False>
-        >>> cb.set_label(('bright_attr', "bar"))
+        >>> cb.set_label(('bright_attr', u"bar"))
         >>> cb
         <CheckBox selectable widget 'bar' state=False>
         """
         self._label.set_text(label)
         # no need to call self._invalidate(). WidgetWrap takes care of
         # that when self.w changes
-    
+
     def get_label(self):
         """
         Return label text.
 
-        >>> cb = CheckBox("Seriously")
-        >>> cb.get_label()
-        'Seriously'
-        >>> cb.label  # Urwid 0.9.9 or later
-        'Seriously'
-        >>> cb.set_label([('bright_attr', "flashy"), " normal"])
-        >>> cb.label  #  only text is returned 
-        'flashy normal'
+        >>> cb = CheckBox(u"Seriously")
+        >>> print cb.get_label()
+        Seriously
+        >>> print cb.label  # Urwid 0.9.9 or later
+        Seriously
+        >>> cb.set_label([('bright_attr', u"flashy"), u" normal"])
+        >>> print cb.label  #  only text is returned
+        flashy normal
         """
         return self._label.text
     label = property(get_label)
-    
+
     def set_state(self, state, do_callback=True):
         """
         Set the CheckBox state.
@@ -232,8 +238,8 @@ class CheckBox(WidgetWrap):
         """
         Toggle state on 'activate' command.  
 
-        >>> assert command_map[' '] == 'activate'
-        >>> assert command_map['enter'] == 'activate'
+        >>> assert CheckBox._command_map[' '] == 'activate'
+        >>> assert CheckBox._command_map['enter'] == 'activate'
         >>> size = (10,)
         >>> cb = CheckBox('press me')
         >>> cb.state
@@ -245,7 +251,7 @@ class CheckBox(WidgetWrap):
         >>> cb.state
         False
         """
-        if command_map[key] != 'activate':
+        if self._command_map[key] != 'activate':
             return key
         
         self.toggle_state()
@@ -314,7 +320,7 @@ class RadioButton(CheckBox):
 
         This function will append the new radio button to group.
         "first True" will set to True if group is empty.
-        
+
         Signals supported: 'change'
         Register signal handler with:
           connect_signal(radio_button, 'change', callback [,user_data])
@@ -323,20 +329,20 @@ class RadioButton(CheckBox):
           disconnect_signal(radio_button, 'change', callback [,user_data])
 
         >>> bgroup = [] # button group
-        >>> b1 = RadioButton(bgroup, "Agree")
-        >>> b2 = RadioButton(bgroup, "Disagree")
+        >>> b1 = RadioButton(bgroup, u"Agree")
+        >>> b2 = RadioButton(bgroup, u"Disagree")
         >>> len(bgroup)
         2
         >>> b1
         <RadioButton selectable widget 'Agree' state=True>
         >>> b2
         <RadioButton selectable widget 'Disagree' state=False>
-        >>> b2.render((15,), focus=True).text  # preview RadioButton
-        ['( ) Disagree   ']
+        >>> b2.render((15,), focus=True).text # ... = b in Python 3
+        [...'( ) Disagree   ']
         """
         if state=="first True":
             state = not group
-        
+
         self.group = group
         self.__super.__init__(label, state, False, on_state_change, 
             user_data)
@@ -355,16 +361,16 @@ class RadioButton(CheckBox):
         group will be set to False.
 
         >>> bgroup = [] # button group
-        >>> b1 = RadioButton(bgroup, "Agree")
-        >>> b2 = RadioButton(bgroup, "Disagree")
-        >>> b3 = RadioButton(bgroup, "Unsure")
+        >>> b1 = RadioButton(bgroup, u"Agree")
+        >>> b2 = RadioButton(bgroup, u"Disagree")
+        >>> b3 = RadioButton(bgroup, u"Unsure")
         >>> b1.state, b2.state, b3.state
         (True, False, False)
         >>> b2.set_state(True)
         >>> b1.state, b2.state, b3.state
         (False, True, False)
         >>> def relabel_button(radio_button, new_state):
-        ...     radio_button.set_label("Think Harder!")
+        ...     radio_button.set_label(u"Think Harder!")
         >>> connect_signal(b3, 'change', relabel_button)
         >>> b3
         <RadioButton selectable widget 'Unsure' state=False>
@@ -428,11 +434,11 @@ class Button(WidgetWrap):
         Unregister signal handlers with:
           disconnect_signal(button, 'click', callback [,user_data])
 
-        >>> Button("Ok")
+        >>> Button(u"Ok")
         <Button selectable widget 'Ok'>
         >>> b = Button("Cancel")
-        >>> b.render((15,), focus=True).text  # preview Button
-        ['< Cancel      >']
+        >>> b.render((15,), focus=True).text # ... = b in Python 3
+        [...'< Cancel      >']
         """
         self._label = SelectableIcon("", 0)    
         cols = Columns([
@@ -440,19 +446,19 @@ class Button(WidgetWrap):
             self._label,
             ('fixed', 1, self.button_right)],
             dividechars=1)
-        self.__super.__init__(cols) 
-        
+        self.__super.__init__(cols)
+
         # The old way of listening for a change was to pass the callback
         # in to the constructor.  Just convert it to the new way:
         if on_press:
             connect_signal(self, 'click', on_press, user_data)
 
         self.set_label(label)
-    
+
     def _repr_words(self):
         # include button.label in repr(button)
         return self.__super._repr_words() + [
-            repr(self.label)]
+            python3_repr(self.label)]
 
     def set_label(self, label):
         """
@@ -461,43 +467,43 @@ class Button(WidgetWrap):
         label -- markup for button label
 
         >>> b = Button("Ok")
-        >>> b.set_label("Yup yup")
+        >>> b.set_label(u"Yup yup")
         >>> b
         <Button selectable widget 'Yup yup'>
         """
         self._label.set_text(label)
-    
+
     def get_label(self):
         """
         Return label text.
 
-        >>> b = Button("Ok")
-        >>> b.get_label()
-        'Ok'
-        >>> b.label  # Urwid 0.9.9 or later
-        'Ok'
+        >>> b = Button(u"Ok")
+        >>> print b.get_label()
+        Ok
+        >>> print b.label  # Urwid 0.9.9 or later
+        Ok
         """
         return self._label.text
     label = property(get_label)
-    
+
     def keypress(self, size, key):
         """
         Send 'click' signal on 'activate' command.
         
-        >>> assert command_map[' '] == 'activate'
-        >>> assert command_map['enter'] == 'activate'
+        >>> assert Button._command_map[' '] == 'activate'
+        >>> assert Button._command_map['enter'] == 'activate'
         >>> size = (15,)
-        >>> b = Button("Cancel")
+        >>> b = Button(u"Cancel")
         >>> clicked_buttons = []
         >>> def handle_click(button):
         ...     clicked_buttons.append(button.label)
         >>> connect_signal(b, 'click', handle_click)
         >>> b.keypress(size, 'enter')
         >>> b.keypress(size, ' ')
-        >>> clicked_buttons
-        ['Cancel', 'Cancel']
+        >>> clicked_buttons # ... = u in Python 2
+        [...'Cancel', ...'Cancel']
         """
-        if command_map[key] != 'activate':
+        if self._command_map[key] != 'activate':
             return key
 
         self._emit('click')
@@ -507,7 +513,7 @@ class Button(WidgetWrap):
         Send 'click' signal on button 1 press.
 
         >>> size = (15,)
-        >>> b = Button("Ok")
+        >>> b = Button(u"Ok")
         >>> clicked_buttons = []
         >>> def handle_click(button):
         ...     clicked_buttons.append(button.label)
@@ -516,16 +522,114 @@ class Button(WidgetWrap):
         True
         >>> b.mouse_event(size, 'mouse press', 2, 4, 0, True) # ignored
         False
-        >>> clicked_buttons
-        ['Ok']
+        >>> clicked_buttons # ... = u in Python 2
+        [...'Ok']
         """
         if button != 1 or not is_mouse_press(event):
             return False
-            
+
         self._emit('click')
         return True
 
 
+class PopUpLauncher(delegate_to_widget_mixin('_original_widget'),
+        WidgetDecoration):
+    def __init__(self, original_widget):
+        self.__super.__init__(original_widget)
+        self._pop_up_widget = None
+
+    def create_pop_up(self):
+        """
+        Subclass must override this method and have is return a widget
+        to be used for the pop-up.  This method is called once each time
+        the pop-up is opened.
+        """
+        raise NotImplementedError("Subclass must override this method")
+
+    def get_pop_up_parameters(self):
+        """
+        Subclass must override this method and have it return a dict, eg:
+
+        {'left':0, 'top':1, 'overlay_width':30, 'overlay_height':4}
+
+        This method is called each time this widget is rendered.
+        """
+        raise NotImplementedError("Subclass must override this method")
+
+    def open_pop_up(self):
+        self._pop_up_widget = self.create_pop_up()
+        self._invalidate()
+
+    def close_pop_up(self):
+        self._pop_up_widget = None
+        self._invalidate()
+
+    def render(self, size, focus=False):
+        canv = self.__super.render(size, focus)
+        if self._pop_up_widget:
+            canv = CompositeCanvas(canv)
+            canv.set_pop_up(self._pop_up_widget, **self.get_pop_up_parameters())
+        return canv
+
+
+class PopUpTarget(WidgetDecoration):
+    # FIXME: this whole class is a terrible hack and must be fixed
+    # when layout and rendering are separated
+    _sizing = set([BOX])
+    _selectable = True
+
+    def __init__(self, original_widget):
+        self.__super.__init__(original_widget)
+        self._pop_up = None
+        self._current_widget = self._original_widget
+
+    def _update_overlay(self, size, focus):
+        canv = self._original_widget.render(size, focus=focus)
+        self._cache_original_canvas = canv # imperfect performance hack
+        pop_up = canv.get_pop_up()
+        if pop_up:
+            left, top, (
+                w, overlay_width, overlay_height) = pop_up
+            if self._pop_up != w:
+                self._pop_up = w
+                self._current_widget = Overlay(w, self._original_widget,
+                    ('fixed left', left), overlay_width,
+                    ('fixed top', top), overlay_height)
+            else:
+                self._current_widget.set_overlay_parameters(
+                    ('fixed left', left), overlay_width,
+                    ('fixed top', top), overlay_height)
+        else:
+            self._pop_up = None
+            self._current_widget = self._original_widget
+
+    def render(self, size, focus=False):
+        self._update_overlay(size, focus)
+        return self._current_widget.render(size, focus=focus)
+    def get_cursor_coords(self, size):
+        self._update_overlay(size, True)
+        return self._current_widget.get_cursor_coords(size)
+    def get_pref_col(self, size):
+        self._update_overlay(size, True)
+        return self._current_widget.get_pref_col(size)
+    def keypress(self, size, key):
+        self._update_overlay(size, True)
+        return self._current_widget.keypress(size, key)
+    def move_cursor_to_coords(self, size, x, y):
+        self._update_overlay(size, True)
+        return self._current_widget.move_cursor_to_coords(size, x, y)
+    def mouse_event(self, size, event, button, x, y, focus):
+        self._update_overlay(size, focus)
+        return self._current_widget.mouse_event(size, event, button, x, y, focus)
+    def pack(self, size=None, focus=False):
+        self._update_overlay(size, focus)
+        return self._current_widget.pack(size)
+
+
+
+
+
+
 def _test():
     import doctest
     doctest.testmod()
